#ifndef INCLUDE_GUARD_FB187AE3_D649_4C2D_B558_E3432B36700D
#define INCLUDE_GUARD_FB187AE3_D649_4C2D_B558_E3432B36700D

#include <Python.h>

namespace cac {

extern PyObject* ca_exception;

extern PyObject* flag_access_rights;
extern PyObject* flag_events;
extern PyObject* enum_type;
extern PyObject* enum_status;
extern PyObject* enum_severity;


}

#endif
